#include <iostream>

#include "snake.h"

#include <mapbox/geometry.hpp>
#include <mapbox/polylabel.hpp>

#include <boost/geometry.hpp>
#include <boost/geometry/geometries/adapted/boost_tuple.hpp>
#include <boost/geometry/geometries/box.hpp>
#include <boost/geometry/geometries/polygon.hpp>

#include "clipper/clipper.hpp"
#define CLIPPER_SCALE 10000

#include "ortools/constraint_solver/routing.h"
#include "ortools/constraint_solver/routing_enums.pb.h"
#include "ortools/constraint_solver/routing_index_manager.h"
#include "ortools/constraint_solver/routing_parameters.h"

using namespace operations_research;

#ifndef NDEBUG
//#define SHOW_TIME
#endif

namespace bg = boost::geometry;
namespace trans = bg::strategy::transform;

BOOST_GEOMETRY_REGISTER_BOOST_TUPLE_CS(bg::cs::cartesian)

namespace snake {
//=========================================================================
// Geometry stuff.
//=========================================================================

void polygonCenter(const BoostPolygon &polygon, BoostPoint &center) {
  using namespace mapbox;
  if (polygon.outer().empty())
    return;
  geometry::polygon<double> p;
  geometry::linear_ring<double> lr1;
  for (size_t i = 0; i < polygon.outer().size(); ++i) {
    geometry::point<double> vertex(polygon.outer()[i].get<0>(),
                                   polygon.outer()[i].get<1>());
    lr1.push_back(vertex);
  }
  p.push_back(lr1);
  geometry::point<double> c = polylabel(p);

  center.set<0>(c.x);
  center.set<1>(c.y);
}

void minimalBoundingBox(const BoostPolygon &polygon, BoundingBox &minBBox) {
  /*
  Find the minimum-area bounding box of a set of 2D points

  The input is a 2D convex hull, in an Nx2 numpy array of x-y co-ordinates.
  The first and last points points must be the same, making a closed polygon.
  This program finds the rotation angles of each edge of the convex polygon,
  then tests the area of a bounding box aligned with the unique angles in
  90 degrees of the 1st Quadrant.
  Returns the

  Tested with Python 2.6.5 on Ubuntu 10.04.4 (original version)
  Results verified using Matlab

  Copyright (c) 2013, David Butterworth, University of Queensland
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

      * Redistributions of source code must retain the above copyright
        notice, this list of conditions and the following disclaimer.
      * Redistributions in binary form must reproduce the above copyright
        notice, this list of conditions and the following disclaimer in the
        documentation and/or other materials provided with the distribution.
      * Neither the name of the Willow Garage, Inc. nor the names of its
        contributors may be used to endorse or promote products derived from
        this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
  */

  if (polygon.outer().empty())
    return;
  BoostPolygon convex_hull;
  bg::convex_hull(polygon, convex_hull);

  // cout << "Convex hull: " << bg::wkt<BoostPolygon2D>(convex_hull) << endl;

  //# Compute edges (x2-x1,y2-y1)
  std::vector<BoostPoint> edges;
  auto convex_hull_outer = convex_hull.outer();
  for (long i = 0; i < long(convex_hull_outer.size()) - 1; ++i) {
    BoostPoint p1 = convex_hull_outer.at(i);
    BoostPoint p2 = convex_hull_outer.at(i + 1);
    double edge_x = p2.get<0>() - p1.get<0>();
    double edge_y = p2.get<1>() - p1.get<1>();
    edges.push_back(BoostPoint{edge_x, edge_y});
  }

  //    cout << "Edges: ";
  //    for (auto e : edges)
  //        cout << e.get<0>() << " " << e.get<1>() << ",";
  //    cout << endl;

  // Calculate unique edge angles  atan2(y/x)
  double angle_scale = 1e3;
  std::set<long> angles_long;
  for (auto vertex : edges) {
    double angle = std::fmod(atan2(vertex.get<1>(), vertex.get<0>()), M_PI / 2);
    angle =
        angle < 0 ? angle + M_PI / 2 : angle; // want strictly positive answers
    angles_long.insert(long(round(angle * angle_scale)));
  }
  std::vector<double> edge_angles;
  for (auto a : angles_long)
    edge_angles.push_back(double(a) / angle_scale);

  //    cout << "Unique angles: ";
  //    for (auto e : edge_angles)
  //        cout << e*180/M_PI << ",";
  //    cout << endl;

  double min_area = std::numeric_limits<double>::infinity();
  // Test each angle to find bounding box with smallest area
  // print "Testing", len(edge_angles), "possible rotations for bounding box...
  // \n"
  for (double angle : edge_angles) {

    trans::rotate_transformer<bg::degree, double, 2, 2> rotate(angle * 180 /
                                                               M_PI);
    BoostPolygon hull_rotated;
    bg::transform(convex_hull, hull_rotated, rotate);
    // cout << "Convex hull rotated: " << bg::wkt<BoostPolygon2D>(hull_rotated)
    // << endl;

    bg::model::box<BoostPoint> box;
    bg::envelope(hull_rotated, box);
    //        cout << "Bounding box: " <<
    //        bg::wkt<bg::model::box<BoostPoint2D>>(box) << endl;

    //# print "Rotated hull points are \n", rot_points
    BoostPoint min_corner = box.min_corner();
    BoostPoint max_corner = box.max_corner();
    double min_x = min_corner.get<0>();
    double max_x = max_corner.get<0>();
    double min_y = min_corner.get<1>();
    double max_y = max_corner.get<1>();
    //        cout << "min_x: " << min_x << endl;
    //        cout << "max_x: " << max_x << endl;
    //        cout << "min_y: " << min_y << endl;
    //        cout << "max_y: " << max_y << endl;

    // Calculate height/width/area of this bounding rectangle
    double width = max_x - min_x;
    double height = max_y - min_y;
    double area = width * height;
    //        cout << "Width: " << width << endl;
    //        cout << "Height: " << height << endl;
    //        cout << "area: " << area << endl;
    //        cout << "angle: " << angle*180/M_PI << endl;

    // Store the smallest rect found first (a simple convex hull might have 2
    // answers with same area)
    if (area < min_area) {
      min_area = area;
      minBBox.angle = angle;
      minBBox.width = width;
      minBBox.height = height;

      minBBox.corners.clear();
      minBBox.corners.outer().push_back(BoostPoint{min_x, min_y});
      minBBox.corners.outer().push_back(BoostPoint{min_x, max_y});
      minBBox.corners.outer().push_back(BoostPoint{max_x, max_y});
      minBBox.corners.outer().push_back(BoostPoint{max_x, min_y});
      minBBox.corners.outer().push_back(BoostPoint{min_x, min_y});
    }
    // cout << endl << endl;
  }

  // Transform corners of minimal bounding box.
  trans::rotate_transformer<bg::degree, double, 2, 2> rotate(-minBBox.angle *
                                                             180 / M_PI);
  BoostPolygon rotated_polygon;
  bg::transform(minBBox.corners, rotated_polygon, rotate);
  minBBox.corners = rotated_polygon;
}

void offsetPolygon(const BoostPolygon &polygon, BoostPolygon &polygonOffset,
                   double offset) {
  bg::strategy::buffer::distance_symmetric<double> distance_strategy(offset);
  bg::strategy::buffer::join_miter join_strategy(3);
  bg::strategy::buffer::end_flat end_strategy;
  bg::strategy::buffer::point_square point_strategy;
  bg::strategy::buffer::side_straight side_strategy;

  bg::model::multi_polygon<BoostPolygon> result;

  bg::buffer(polygon, result, distance_strategy, side_strategy, join_strategy,
             end_strategy, point_strategy);

  if (result.size() > 0)
    polygonOffset = result[0];
}

void graphFromPolygon(const BoostPolygon &polygon,
                      const BoostLineString &vertices, Matrix<double> &graph) {
  size_t n = graph.getN();

  for (size_t i = 0; i < n; ++i) {
    BoostPoint v1 = vertices[i];
    for (size_t j = i + 1; j < n; ++j) {
      BoostPoint v2 = vertices[j];
      BoostLineString path{v1, v2};

      double distance = 0;
      if (!bg::within(path, polygon))
        distance = std::numeric_limits<double>::infinity();
      else
        distance = bg::length(path);

      graph.set(i, j, distance);
      graph.set(j, i, distance);
    }
  }
}

bool dijkstraAlgorithm(
    const size_t numElements, size_t startIndex, size_t endIndex,
    std::vector<size_t> &elementPath,
    std::function<double(const size_t, const size_t)> distanceDij) {
  if (startIndex >= numElements || endIndex >= numElements ||
      endIndex == startIndex) {
    return false;
  }
  // Node struct
  // predecessorIndex is the index of the predecessor node
  // (nodeList[predecessorIndex]) distance is the distance between the node and
  // the start node node number is stored by the position in nodeList
  struct Node {
    int predecessorIndex = -1;
    double distance = std::numeric_limits<double>::infinity();
  };

  // The list with all Nodes (elements)
  std::vector<Node> nodeList(numElements);
  // This list will be initalized with indices referring to the elements of
  // nodeList. Elements will be successively remove during the execution of the
  // Dijkstra Algorithm.
  std::vector<size_t> workingSet(numElements);

  // append elements to node list
  for (size_t i = 0; i < numElements; ++i)
    workingSet[i] = i;

  nodeList[startIndex].distance = 0;

  // Dijkstra Algorithm
  // https://de.wikipedia.org/wiki/Dijkstra-Algorithmus
  while (workingSet.size() > 0) {
    // serach Node with minimal distance
    double minDist = std::numeric_limits<double>::infinity();
    int minDistIndex_WS = -1; // WS = workinSet
    for (size_t i = 0; i < workingSet.size(); ++i) {
      const int nodeIndex = workingSet.at(i);
      const double dist = nodeList.at(nodeIndex).distance;
      if (dist < minDist) {
        minDist = dist;
        minDistIndex_WS = i;
      }
    }
    if (minDistIndex_WS == -1)
      return false;

    size_t indexU_NL = workingSet.at(minDistIndex_WS); // NL = nodeList
    workingSet.erase(workingSet.begin() + minDistIndex_WS);
    if (indexU_NL == endIndex) // shortest path found
      break;

    const double distanceU = nodeList.at(indexU_NL).distance;
    // update distance
    for (size_t i = 0; i < workingSet.size(); ++i) {
      int indexV_NL = workingSet[i]; // NL = nodeList
      Node *v = &nodeList[indexV_NL];
      double dist = distanceDij(indexU_NL, indexV_NL);
      // is ther an alternative path which is shorter?
      double alternative = distanceU + dist;
      if (alternative < v->distance) {
        v->distance = alternative;
        v->predecessorIndex = indexU_NL;
      }
    }
  }
  // end Djikstra Algorithm

  // reverse assemble path
  int e = endIndex;
  while (1) {
    if (e == -1) {
      if (elementPath[0] == startIndex) // check if starting point was reached
        break;
      return false;
    }
    elementPath.insert(elementPath.begin(), e);

    // Update Node
    e = nodeList[e].predecessorIndex;
  }
  return true;
}

void toDistanceMatrix(Matrix<double> &graph) {
  size_t n = graph.getN();

  auto distance = [graph](size_t i, size_t j) { return graph.get(i, j); };

  std::vector<size_t> path;
  for (size_t i = 0; i < n; ++i) {
    for (size_t j = i + 1; j < n; ++j) {
      double d = graph.get(i, j);
      if (!std::isinf(d))
        continue;
      path.clear();
      bool ret = dijkstraAlgorithm(n, i, j, path, distance);
      assert(ret);
      (void)ret;
      //            cout << "(" << i << "," << j << ") d: " << d << endl;
      //            cout << "Path size: " << path.size() << endl;
      //            for (auto idx : path)
      //                cout << idx << " ";
      //            cout << endl;

      d = 0;
      for (long k = 0; k < long(path.size()) - 1; ++k) {
        size_t idx0 = path[k];
        size_t idx1 = path[k + 1];
        double d0 = graph.get(idx0, idx1);
        assert(std::isinf(d0) == false);
        d += d0;
      }

      graph.set(i, j, d);
      graph.set(j, i, d);
    }
  }
}

void shortestPathFromGraph(const Matrix<double> &graph, size_t startIndex,
                           size_t endIndex, std::vector<size_t> &pathIdx) {

  if (!std::isinf(graph.get(startIndex, endIndex))) {
    pathIdx.push_back(startIndex);
    pathIdx.push_back(endIndex);
  } else {
    auto distance = [graph](size_t i, size_t j) { return graph.get(i, j); };
    bool ret = dijkstraAlgorithm(graph.getN(), startIndex, endIndex, pathIdx,
                                 distance);
    assert(ret);
    (void)ret;
  }

  //=========================================================================
  // Scenario calculation.
  //=========================================================================
}
Scenario::Scenario()
    : _tileWidth(5 * bu::si::meter), _tileHeight(5 * bu::si::meter),
      _minTileArea(0 * bu::si::meter * bu::si::meter), _needsUpdate(true) {}

void Scenario::setMeasurementArea(const BoostPolygon &area) {
  _needsUpdate = true;
  _mArea = area;
}

void Scenario::setServiceArea(const BoostPolygon &area) {
  _needsUpdate = true;
  _sArea = area;
}

void Scenario::setCorridor(const BoostPolygon &area) {
  _needsUpdate = true;
  _corridor = area;
}

BoostPolygon &Scenario::measurementArea() {
  _needsUpdate = true;
  return _mArea;
}

BoostPolygon &Scenario::serviceArea() {
  _needsUpdate = true;
  return _sArea;
}

BoostPolygon &Scenario::corridor() {
  _needsUpdate = true;
  return _corridor;
}

const BoundingBox &Scenario::mAreaBoundingBox() const {
  return _mAreaBoundingBox;
}

const BoostPolygon &Scenario::measurementArea() const { return _mArea; }

const BoostPolygon &Scenario::serviceArea() const { return _sArea; }

const BoostPolygon &Scenario::corridor() const { return _corridor; }

const BoostPolygon &Scenario::joinedArea() const { return _jArea; }

const vector<BoostPolygon> &Scenario::tiles() const { return _tiles; }

const BoostLineString &Scenario::tileCenterPoints() const {
  return _tileCenterPoints;
}

const BoundingBox &Scenario::measurementAreaBBox() const {
  return _mAreaBoundingBox;
}

const BoostPoint &Scenario::homePositon() const { return _homePosition; }

bool Scenario::update() {
  if (!_needsUpdate)
    return true;
  if (!_calculateBoundingBox())
    return false;
  if (!_calculateTiles())
    return false;
  if (!_calculateJoinedArea())
    return false;
  _needsUpdate = false;
  return true;
}

bool Scenario::_calculateBoundingBox() {
  minimalBoundingBox(_mArea, _mAreaBoundingBox);
  return true;
}

/**
 * Devides the (measurement area) bounding  box into tiles and clips it to the
 * measurement area.
 *
 * Devides the (measurement area) bounding  box into tiles of width \p tileWidth
 * and height \p tileHeight. Clips the resulting tiles to the measurement area.
 * Tiles are rejected, if their area is smaller than \p minTileArea. The
 * function assumes that \a _mArea and \a _mAreaBoundingBox have correct values.
 * \see \ref Scenario::_areas2enu() and \ref Scenario::_calculateBoundingBox().
 *
 * @param tileWidth The width (>0) of a tile.
 * @param tileHeight The heigth (>0) of a tile.
 * @param minTileArea The minimal area (>0) of a tile.
 *
 * @return Returns true if successful.
 */
bool Scenario::_calculateTiles() {
  _tiles.clear();
  _tileCenterPoints.clear();

  if (_tileWidth <= 0 * bu::si::meter || _tileHeight <= 0 * bu::si::meter ||
      _minTileArea <= 0 * bu::si::meter * bu::si::meter) {
    errorString =
        "Parameters tileWidth, tileHeight, minTileArea must be positive.";
    return false;
  }

  double bboxWidth = _mAreaBoundingBox.width;
  double bboxHeight = _mAreaBoundingBox.height;

  BoostPoint origin = _mAreaBoundingBox.corners.outer()[0];

  // cout << "Origin: " << origin[0] << " " << origin[1] << endl;
  // Transform _mArea polygon to bounding box coordinate system.
  trans::rotate_transformer<boost::geometry::degree, double, 2, 2> rotate(
      _mAreaBoundingBox.angle * 180 / M_PI);
  trans::translate_transformer<double, 2, 2> translate(-origin.get<0>(),
                                                       -origin.get<1>());
  BoostPolygon translated_polygon;
  BoostPolygon rotated_polygon;
  boost::geometry::transform(_mArea, translated_polygon, translate);
  boost::geometry::transform(translated_polygon, rotated_polygon, rotate);
  bg::correct(rotated_polygon);

  // cout << bg::wkt<BoostPolygon2D>(rotated_polygon) << endl;

  size_t iMax = ceil(bboxWidth / _tileWidth.value());
  size_t jMax = ceil(bboxHeight / _tileHeight.value());

  if (iMax < 1 || jMax < 1) {
    std::stringstream ss;
    ss << "Tile width or Tile height to large. "
       << "tile width = " << _tileWidth << ", "
       << "tile height = " << _tileHeight << "." << std::endl;
    errorString = ss.str();
    return false;
  }

  trans::rotate_transformer<boost::geometry::degree, double, 2, 2> rotate_back(
      -_mAreaBoundingBox.angle * 180 / M_PI);
  trans::translate_transformer<double, 2, 2> translate_back(origin.get<0>(),
                                                            origin.get<1>());
  for (size_t i = 0; i < iMax; ++i) {
    double x_min = _tileWidth.value() * i;
    double x_max = x_min + _tileWidth.value();
    for (size_t j = 0; j < jMax; ++j) {
      double y_min = _tileHeight.value() * j;
      double y_max = y_min + _tileHeight.value();

      BoostPolygon tile_unclipped;
      tile_unclipped.outer().push_back(BoostPoint{x_min, y_min});
      tile_unclipped.outer().push_back(BoostPoint{x_min, y_max});
      tile_unclipped.outer().push_back(BoostPoint{x_max, y_max});
      tile_unclipped.outer().push_back(BoostPoint{x_max, y_min});
      tile_unclipped.outer().push_back(BoostPoint{x_min, y_min});

      std::deque<BoostPolygon> boost_tiles;
      if (!boost::geometry::intersection(tile_unclipped, rotated_polygon,
                                         boost_tiles))
        continue;

      for (BoostPolygon t : boost_tiles) {
        if (bg::area(t) > _minTileArea.value()) {
          // Transform boost_tile to world coordinate system.
          BoostPolygon rotated_tile;
          BoostPolygon translated_tile;
          boost::geometry::transform(t, rotated_tile, rotate_back);
          boost::geometry::transform(rotated_tile, translated_tile,
                                     translate_back);

          // Store tile and calculate center point.
          _tiles.push_back(translated_tile);
          BoostPoint tile_center;
          polygonCenter(translated_tile, tile_center);
          _tileCenterPoints.push_back(tile_center);
        }
      }
    }
  }

  if (_tiles.size() < 1) {
    errorString =
        "No tiles calculated. Is the minTileArea parameter large enough?";
    return false;
  }

  return true;
}

bool Scenario::_calculateJoinedArea() {
  _jArea.clear();
  // Measurement area and service area overlapping?
  bool overlapingSerMeas = bg::intersects(_mArea, _sArea) ? true : false;
  bool corridorValid = _corridor.outer().size() > 0 ? true : false;

  // Check if corridor is connecting measurement area and service area.
  bool corridor_is_connection = false;
  if (corridorValid) {
    // Corridor overlaping with measurement area?
    if (bg::intersects(_corridor, _mArea)) {
      // Corridor overlaping with service area?
      if (bg::intersects(_corridor, _sArea)) {
        corridor_is_connection = true;
      }
    }
  }

  // Are areas joinable?
  std::deque<BoostPolygon> sol;
  BoostPolygon partialArea = _mArea;
  if (overlapingSerMeas) {
    if (corridor_is_connection) {
      bg::union_(partialArea, _corridor, sol);
    }
  } else if (corridor_is_connection) {
    bg::union_(partialArea, _corridor, sol);
  } else {
    errorString = "Areas are not overlapping";
    return false;
  }

  if (sol.size() > 0) {
    partialArea = sol[0];
    sol.clear();
  }

  // Join areas.
  bg::union_(partialArea, _sArea, sol);

  if (sol.size() > 0) {
    _jArea = sol[0];
  } else {
    return false;
  }

  return true;
}

Area Scenario::minTileArea() const { return _minTileArea; }

void Scenario::setMinTileArea(Area minTileArea) {
  if (minTileArea >= 0 * bu::si::meter * bu::si::meter) {
    _needsUpdate = true;
    _minTileArea = minTileArea;
  }
}

Length Scenario::tileHeight() const { return _tileHeight; }

void Scenario::setTileHeight(Length tileHeight) {
  if (tileHeight > 0 * bu::si::meter) {
    _needsUpdate = true;
    _tileHeight = tileHeight;
  }
}

Length Scenario::tileWidth() const { return _tileWidth; }

void Scenario::setTileWidth(Length tileWidth) {
  if (tileWidth > 0 * bu::si::meter) {
    _needsUpdate = true;
    _tileWidth = tileWidth;
  }
}

//=========================================================================
// Tile calculation.
//=========================================================================

bool joinAreas(const std::vector<BoostPolygon> &areas,
               BoostPolygon &joinedArea) {
  if (areas.size() < 1)
    return false;
  std::deque<std::size_t> idxList;
  for (size_t i = 1; i < areas.size(); ++i)
    idxList.push_back(i);
  joinedArea = areas[0];

  std::deque<BoostPolygon> sol;
  while (idxList.size() > 0) {
    bool success = false;
    for (auto it = idxList.begin(); it != idxList.end(); ++it) {
      bg::union_(joinedArea, areas[*it], sol);
      if (sol.size() > 0) {
        joinedArea = sol[0];
        sol.clear();
        idxList.erase(it);
        success = true;
        break;
      }
    }
    if (!success)
      return false;
  }

  return true;
}

BoundingBox::BoundingBox() : width(0), height(0), angle(0) {}

void BoundingBox::clear() {
  width = 0;
  height = 0;
  angle = 0;
  corners.clear();
}

bool flight_plan::transectsFromScenario(Length distance, Length minLength,
                                        Angle angle, const Scenario &scenario,
                                        const Progress &p,
                                        flight_plan::Transects &t,
                                        string &errorString) {
  // Rotate measurement area by angle and calculate bounding box.
  auto &mArea = scenario.measurementArea();
  BoostPolygon mAreaRotated;
  trans::rotate_transformer<bg::degree, double, 2, 2> rotate(-angle.value() *
                                                             180 / M_PI);
  bg::transform(mArea, mAreaRotated, rotate);

  BoostBox box;
  boost::geometry::envelope(mAreaRotated, box);

  double x0 = box.min_corner().get<0>();
  double y0 = box.min_corner().get<1>();
  double x1 = box.max_corner().get<0>();
  double y1 = box.max_corner().get<1>();

  // Generate transects and convert them to clipper path.
  size_t num_t = int(ceil((y1 - y0) / distance.value())); // number of transects
  trans::rotate_transformer<bg::degree, double, 2, 2> rotate_back(
      angle.value() * 180 / M_PI);
  vector<ClipperLib::Path> transectsClipper;
  transectsClipper.reserve(num_t);
  for (size_t i = 0; i < num_t; ++i) {
    // calculate transect
    BoostPoint v1{x0, y0 + i * distance.value()};
    BoostPoint v2{x1, y0 + i * distance.value()};
    BoostLineString transect;
    transect.push_back(v1);
    transect.push_back(v2);

    // transform back
    BoostLineString temp_transect;
    bg::transform(transect, temp_transect, rotate_back);

    ClipperLib::IntPoint c1{
        static_cast<ClipperLib::cInt>(transect[0].get<0>() * CLIPPER_SCALE),
        static_cast<ClipperLib::cInt>(transect[0].get<1>() * CLIPPER_SCALE)};
    ClipperLib::IntPoint c2{
        static_cast<ClipperLib::cInt>(transect[1].get<0>() * CLIPPER_SCALE),
        static_cast<ClipperLib::cInt>(transect[1].get<1>() * CLIPPER_SCALE)};
    ClipperLib::Path path{c1, c2};
    transectsClipper.push_back(path);
  }

  if (transectsClipper.size() == 0) {
    std::stringstream ss;
    ss << "Not able to generate transects. Parameter: distance = " << distance
       << std::endl;
    errorString = ss.str();
    return false;
  }

  // Convert measurement area to clipper path.
  ClipperLib::Path mAreaClipper;
  for (auto vertex : mArea.outer()) {
    mAreaClipper.push_back(ClipperLib::IntPoint{
        static_cast<ClipperLib::cInt>(vertex.get<0>() * CLIPPER_SCALE),
        static_cast<ClipperLib::cInt>(vertex.get<1>() * CLIPPER_SCALE)});
  }

  // Perform clipping.
  // Clip transects to measurement area.
  ClipperLib::Clipper clipper;
  clipper.AddPath(mAreaClipper, ClipperLib::ptClip, true);
  clipper.AddPaths(transectsClipper, ClipperLib::ptSubject, false);
  ClipperLib::PolyTree clippedTransecs;
  clipper.Execute(ClipperLib::ctIntersection, clippedTransecs,
                  ClipperLib::pftNonZero, ClipperLib::pftNonZero);
  auto &transects = clippedTransecs;

  bool ignoreProgress = p.size() != scenario.tiles().size();
  ClipperLib::PolyTree clippedTransecs2;
  if (!ignoreProgress) {
    // Calculate processed tiles (_progress[i] == 100) and subtract them from
    // measurement area.
    size_t numTiles = p.size();
    vector<BoostPolygon> processedTiles;
    const auto &tiles = scenario.tiles();
    for (size_t i = 0; i < numTiles; ++i) {
      if (p[i] == 100) {
        processedTiles.push_back(tiles[i]);
      }
    }

    if (processedTiles.size() != numTiles) {
      vector<ClipperLib::Path> processedTilesClipper;
      for (auto t : processedTiles) {
        ClipperLib::Path path;
        for (auto vertex : t.outer()) {
          path.push_back(ClipperLib::IntPoint{
              static_cast<ClipperLib::cInt>(vertex.get<0>() * CLIPPER_SCALE),
              static_cast<ClipperLib::cInt>(vertex.get<1>() * CLIPPER_SCALE)});
        }
        processedTilesClipper.push_back(path);
      }

      // Subtract holes (tiles with measurement_progress == 100) from transects.
      clipper.Clear();
      for (auto &child : clippedTransecs.Childs)
        clipper.AddPath(child->Contour, ClipperLib::ptSubject, false);
      clipper.AddPaths(processedTilesClipper, ClipperLib::ptClip, true);
      clipper.Execute(ClipperLib::ctDifference, clippedTransecs2,
                      ClipperLib::pftNonZero, ClipperLib::pftNonZero);
      transects = clippedTransecs2;
    }
  }

  // Extract transects from  PolyTree and convert them to BoostLineString
  for (auto &child : transects.Childs) {
    auto &clipperTransect = child->Contour;
    BoostPoint v1{static_cast<double>(clipperTransect[0].X) / CLIPPER_SCALE,
                  static_cast<double>(clipperTransect[0].Y) / CLIPPER_SCALE};
    BoostPoint v2{static_cast<double>(clipperTransect[1].X) / CLIPPER_SCALE,
                  static_cast<double>(clipperTransect[1].Y) / CLIPPER_SCALE};

    BoostLineString transect{v1, v2};
    if (bg::length(transect) >= minLength.value())
      t.push_back(transect);
  }

  if (t.size() == 0) {
    std::stringstream ss;
    ss << "Not able to generate transects. Parameter: minLength = " << minLength
       << std::endl;
    errorString = ss.str();
    return false;
  }
  return true;
}

struct RoutingDataModel {
  Matrix<int64_t> distanceMatrix;
  long numVehicles;
  RoutingIndexManager::NodeIndex depot;
};

void generateRoutingModel(const BoostLineString &vertices,
                          const BoostPolygon &polygonOffset, size_t n0,
                          RoutingDataModel &dataModel, Matrix<double> &graph) {

#ifdef SHOW_TIME
  auto start = std::chrono::high_resolution_clock::now();
#endif
  graphFromPolygon(polygonOffset, vertices, graph);
#ifdef SHOW_TIME
  auto delta = std::chrono::duration_cast<std::chrono::milliseconds>(
      std::chrono::high_resolution_clock::now() - start);
  cout << "Execution time graphFromPolygon(): " << delta.count() << " ms"
       << endl;
#endif
  Matrix<double> distanceMatrix(graph);
#ifdef SHOW_TIME
  start = std::chrono::high_resolution_clock::now();
#endif
  toDistanceMatrix(distanceMatrix);
#ifdef SHOW_TIME
  delta = std::chrono::duration_cast<std::chrono::milliseconds>(
      std::chrono::high_resolution_clock::now() - start);
  cout << "Execution time toDistanceMatrix(): " << delta.count() << " ms"
       << endl;
#endif

  dataModel.distanceMatrix.setDimension(n0, n0);
  for (size_t i = 0; i < n0; ++i) {
    dataModel.distanceMatrix.set(i, i, 0);
    for (size_t j = i + 1; j < n0; ++j) {
      dataModel.distanceMatrix.set(
          i, j, int64_t(distanceMatrix.get(i, j) * CLIPPER_SCALE));
      dataModel.distanceMatrix.set(
          j, i, int64_t(distanceMatrix.get(i, j) * CLIPPER_SCALE));
    }
  }
  dataModel.numVehicles = 1;
  dataModel.depot = 0;
}

bool flight_plan::route(const BoostPolygon &area,
                        const flight_plan::Transects &transects,
                        Transects &transectsRouted, flight_plan::Route &route,
                        string &errorString) {
  //=======================================
  // Route Transects using Google or-tools.
  //=======================================

  // Create vertex list;
  BoostLineString vertices;
  size_t n0 = 0;
  for (const auto &t : transects) {
    if (t.size() >= 2) {
      n0 += 2;
    } else {
      n0 += 1;
    }
  }
  vertices.reserve(n0);
  struct TransectInfo {
    TransectInfo(size_t n, bool f) : index(n), front(f) {}

    size_t index;
    bool front;
  };
  std::vector<TransectInfo> transectInfoList;
  size_t idx = 0;
  for (size_t i = 0; i < transects.size(); ++i) {
    const auto &t = transects[i];
    vertices.push_back(t.front());
    transectInfoList.push_back(TransectInfo{i, true});
    if (t.size() >= 2) {
      vertices.push_back(t.back());
      transectInfoList.push_back(TransectInfo{i, false});
    }
  }

  for (long i = 0; i < long(area.outer().size()) - 1; ++i) {
    vertices.push_back(area.outer()[i]);
  }
  for (auto &ring : area.inners()) {
    for (auto &vertex : ring)
      vertices.push_back(vertex);
  }
  size_t n1 = vertices.size();
  // Generate routing model.
  RoutingDataModel dataModel;
  Matrix<double> connectionGraph(n1, n1);
  // Offset joined area.
  BoostPolygon areaOffset;
  offsetPolygon(area, areaOffset, detail::offsetConstant);
#ifdef SHOW_TIME
  start = std::chrono::high_resolution_clock::now();
#endif
  generateRoutingModel(vertices, areaOffset, n0, dataModel, connectionGraph);
#ifdef SHOW_TIME
  delta = std::chrono::duration_cast<std::chrono::milliseconds>(
      std::chrono::high_resolution_clock::now() - start);
  cout << "Execution time _generateRoutingModel(): " << delta.count() << " ms"
       << endl;
#endif

  // Create Routing Index Manager.
  RoutingIndexManager manager(dataModel.distanceMatrix.getN(),
                              dataModel.numVehicles, dataModel.depot);
  // Create Routing Model.
  RoutingModel routing(manager);

  // Create and register a transit callback.
  const int transitCallbackIndex = routing.RegisterTransitCallback(
      [&dataModel, &manager](int64 from_index, int64 to_index) -> int64 {
        // Convert from routing variable Index to distance matrix NodeIndex.
        auto from_node = manager.IndexToNode(from_index).value();
        auto to_node = manager.IndexToNode(to_index).value();
        return dataModel.distanceMatrix.get(from_node, to_node);
      });

  // Define cost of each arc.
  routing.SetArcCostEvaluatorOfAllVehicles(transitCallbackIndex);

  // Define Constraints (this constraints have a huge impact on the
  // solving time, improvments could be done, e.g. SearchFilter).
  Solver *solver = routing.solver();
  size_t index = 0;
  for (size_t i = 0; i < transects.size(); ++i) {
    const auto &t = transects[i];
    if (t.size() >= 2) {
      auto idx0 = manager.NodeToIndex(RoutingIndexManager::NodeIndex(index));
      auto idx1 =
          manager.NodeToIndex(RoutingIndexManager::NodeIndex(index + 1));
      auto cond0 = routing.NextVar(idx0)->IsEqual(idx1);
      auto cond1 = routing.NextVar(idx1)->IsEqual(idx0);
      vector<IntVar *> conds{cond0, cond1};
      auto c = solver->MakeAllDifferent(conds);
      solver->MakeRejectFilter();
      solver->AddConstraint(c);
      index += 2;
    } else {
      index += 1;
    }
  }

  // Setting first solution heuristic.
  RoutingSearchParameters searchParameters = DefaultRoutingSearchParameters();
  searchParameters.set_first_solution_strategy(
      FirstSolutionStrategy::PATH_CHEAPEST_ARC);
  google::protobuf::Duration *tMax =
      new google::protobuf::Duration(); // seconds
  tMax->set_seconds(10);
  searchParameters.set_allocated_time_limit(tMax);

  // Solve the problem.
#ifdef SHOW_TIME
  start = std::chrono::high_resolution_clock::now();
#endif
  const Assignment *solution = routing.SolveWithParameters(searchParameters);
#ifdef SHOW_TIME
  delta = std::chrono::duration_cast<std::chrono::milliseconds>(
      std::chrono::high_resolution_clock::now() - start);
  cout << "Execution time routing.SolveWithParameters(): " << delta.count()
       << " ms" << endl;
#endif

  if (!solution || solution->Size() <= 1) {
    errorString = "Not able to solve the routing problem.";
    return false;
  }

  // Extract index list from solution.
  index = routing.Start(0);
  std::vector<size_t> route_idx;
  route_idx.push_back(manager.IndexToNode(index).value());
  while (!routing.IsEnd(index)) {
    index = solution->Value(routing.NextVar(index));
    route_idx.push_back(manager.IndexToNode(index).value());
  }

  // Helper Lambda.
  auto idx2Vertex = [&vertices](const std::vector<size_t> &idxArray,
                                std::vector<BoostPoint> &path) {
    for (auto idx : idxArray)
      path.push_back(vertices[idx]);
  };

  // Construct route.
  for (size_t i = 0; i < route_idx.size() - 1; ++i) {
    size_t idx0 = route_idx[i];
    size_t idx1 = route_idx[i + 1];
    const auto &info1 = transectInfoList[idx0];
    const auto &info2 = transectInfoList[idx1];
    if (info1.index == info2.index) { // same transect?
      if (!info1.front) {             // transect reversal needed?
        BoostLineString reversedTransect;
        const auto &t = transects[info1.index];
        for (auto it = t.end() - 1; it >= t.begin(); --it) {
          reversedTransect.push_back(*it);
        }
        transectsRouted.push_back(reversedTransect);
        for (auto it = reversedTransect.begin();
             it < reversedTransect.end() - 1; ++it) {
          route.push_back(*it);
        }
      } else {
        const auto &t = transects[info1.index];
        for (auto it = t.begin(); it < t.end() - 1; ++it) {
          route.push_back(*it);
        }
        transectsRouted.push_back(t);
      }
    } else {
      std::vector<size_t> idxList;
      shortestPathFromGraph(connectionGraph, idx0, idx1, idxList);
      if (i != route_idx.size() - 2) {
        idxList.pop_back();
      }
      idx2Vertex(idxList, route);
    }
  }
}

} // namespace snake

bool boost::geometry::model::operator==(snake::BoostPoint p1,
                                        snake::BoostPoint p2) {
  return (p1.get<0>() == p2.get<0>()) && (p1.get<1>() == p2.get<1>());
}

bool boost::geometry::model::operator!=(snake::BoostPoint p1,
                                        snake::BoostPoint p2) {
  return !(p1 == p2);
}
